#define TRIGGER 20 // Trigger level to detect a strike.
#define MAXPADS 4 // number of pads
#define SAMPLES 16 // number of samples to read over. (Should be less than 16).
                   // 10~=0.5ms latency
#define RELEASE_SAMPLES 1300 // number of samples to wait before resetting trigger
                             // 1000 = 52ms.
#define BASESAMPLES 20 // number of samples to take when starting up
                       // to read base level. (should be less than 32)
#define SERIALRATE 230400 // Serial port rate. (yes you can set it that high)
#define STARTUP_DELAY 2000 // time to wait for the capacitor on the reference line to charge.
//#define DEBUG // whether output should be human readable.

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define STATE_WAIT 0
#define STATE_READ 1
#define STATE_RELEASE 2

unsigned int samples=0;
char piezoIn[]={0,1,2,3}; // analogue pin numbers for each pad.
int piezoVal[]={0,0,0,0};
int piezoSum[]={0,0,0,0}; // accumulated sum of readings for each sensor
int piezoBaseVal[]={0,0,0,0}; // base level values of each sensor voltage
char state=STATE_WAIT; // what's happening
  // 0= waiting for strike, 1=reading strike data, 2= waiting for RELEASESAMPLES samples to avoid retriggering.
int baseSum; // used when calculating base levels.

void setup()
{
  // set analog sample rate to 76 kHz (clock prescaler = 16)
  cbi(ADCSRA,ADPS0);
  cbi(ADCSRA,ADPS1);
  sbi(ADCSRA,ADPS2);
  // set up inputs.
  for (char i=0; i<MAXPADS; i++) {
    pinMode(piezoIn[i],INPUT); // set pin modes to analogue in.
  }
  Serial.begin(SERIALRATE);
  analogReference(DEFAULT);
  delay(STARTUP_DELAY); // allow the capacitor on the 2.5v reference line to charge.
  for (char i=0; i<MAXPADS; i++) {
    baseSum=0;
    for (char j=0; j<BASESAMPLES; j++) {
      baseSum+=analogRead(piezoIn[i]); // read base level for each sensor
    }
    piezoBaseVal[i]=baseSum/BASESAMPLES;
  }
  #ifdef DEBUG // print out the base level if DEBUG is set.
  Serial.print("Base level: ");
  for (char i=0; i<MAXPADS; i++) {
    if (i>0) Serial.print (", ");
    Serial.print(piezoBaseVal[i]);
  }
  Serial.print("\n");
  #endif
}

void printStrike() {
  // print the data from a single strike to the serial port.
  #ifdef DEBUG // print human readable values if DEBUG is set
  long total;
  Serial.print("Raw: ");
  total=0;
  for (char i=0; i<MAXPADS; i++) {
    if (i > 0) { Serial.print(", "); };
    Serial.print(piezoSum[i]); // print summed absolute values for each sensor.
    total+=piezoSum[i];
  }
  Serial.print("; Total: ");
  Serial.print(total/SAMPLES);
  Serial.print("; Sensor percent: ");
  for (char i=0; i<MAXPADS; i++) {
    if (i > 0) { Serial.print(", "); };
    Serial.print((piezoSum[i]*100)/(total));
  }
  Serial.print("\n");
  #endif
  #ifndef DEBUG // if DEBUG is off, print machine readable values
  unsigned int valOut; 
  unsigned char cOut;
  for (char i=0; i<MAXPADS; i++) {
    valOut=piezoSum[i];
    // put the output values into the lower 7 bits of 2 bytes, high part first.
    cOut=(valOut >> 7) & 0x7f;
    Serial.write(cOut);
    cOut=valOut & 0x7f;
    if (i==3) {
      cOut|=0x80; // last character has the high bit set as a flag.
    }
    Serial.write(cOut);
  }
  #endif
}


void loop()
{
  int val;
  // read values on all the sensors.
  for (char i=0; i<MAXPADS; i++) {
    val=analogRead(piezoIn[i]); // read sensor value.
    piezoVal[i]=abs(val-piezoBaseVal[i]);
  }
  samples++;
  switch (state) {
    case STATE_WAIT: // waiting for a strike
      for (char i=0; i<MAXPADS; i++) {
        if (piezoVal[i]>TRIGGER) {
          // trigger level has been reached, so go to STATE_READ.
          state=STATE_READ;
          samples=0;
          // reset all the sums.
          for (char j=0; j<MAXPADS; j++) {
            piezoSum[j]=0;
          }
          break; // no need to read the other sensors.
        }
      }
      break;
    case STATE_READ: // reading strike signals.
      for (char i=0; i<MAXPADS; i++){
        piezoSum[i]+=piezoVal[i];
      }
      if (samples >= SAMPLES) {
        // the strike has been read, so send the result and go to STATE_RELEASE.
        printStrike();
        state=STATE_RELEASE;
        samples=0; // reset sample counter.
      }
      break;
    case STATE_RELEASE: // delay while vibrations from the strike die down.
      if (samples >= RELEASE_SAMPLES) {
        state=STATE_WAIT;
      }
  }
}
