#!/usr/bin/python

import sys,os,time
from pyseq import * # (from http://www.sci.ccny.cuny.edu/~brinkman/software/pyseq/)
import threading
from numpy import * # numerical python library
from scipy.optimize import leastsq # least squares algorithm from scientific python library
import serial

# Init variables.
usbPort="/dev/ttyUSB0" # usb serial device to use
usbBaud=230400
calFile="mmap.cal" # name of file for calibration data (same dir as the program)
calDumpFile="mmap.dump" # name of file for raw calib data
velOutRange=[31,127] # min and max midi velocities to send
midich=0 # midi channel to use
# Modes:
MODE_STOP=0
MODE_MIDI=1
MODE_CAL=2
MODE_QUIT=3
# Other global vars.

# Defines the equation which is fitted to the raw data.
# p = array of coefficients
# v = velocity value
# s1..s4 = sensor readings
def mapCurve(p,s1,s2,s3,s4):
   #print "coeffs:", p
   k2,k3,k4,l1,l2,l3,l4=p # give names to the coefficients.
   f1=s1 # first coefficient is always 1 
   f2=s2*k2
   f3=s3*k3 
   f4=s4*k4
   v=f1+f2+f3+f4
   x=(l1*f1 +
      l2*f2 +
      l3*f3 +
      l4*f4)/v
   return x

# Puts the equation in mapCurve into a form suitable to be passed to the leastsq function.
def mapCurveErr(p,val,s1,s2,s3,s4):
   err=val-mapCurve(p,s1,s2,s3,s4) # leastsq needs a function which returns an error value
      # i.e. the difference between the calculated and actual x or y value.
   return err

def velCurve(p,vr,x,y):
   # curve to fit to velocity data.
   # vr=raw velocity
   # x,y=mapped coordinates
   # returns a corrected velocity value scaled to the range [0,1]
   k1,k2,k3,k4,k5,k6,k7,k8,k9,k10,k11,k12=p # give names to the coefficients.
   v=k1+k2*x+k3*y+k4*x*x+k5*y*y+k11*x*y+ \
     vr*(k6+k7*x+k8*y+k9*x*x+k10*y*y+k12*x*y)
   return v

def velCurveErr(p,vel,rawvel,x,y):
   # error function for the velocity mapping.
   err=vel-velCurve(p,rawvel,x,y) # leastsq needs a function which returns an error value
      # i.e. the difference between the calculated and actual velocity value.
   return err

# The class which deals with reading sensor values from the arduino's serial port.
class PortReader(threading.Thread):

   # class initialisation code
   def __init__(self,usbPort):
      threading.Thread.__init__(self)
      self.usbPort=usbPort
      self.mode=MODE_STOP
      self.quit=False
      self.results=[]

   # method to return an array of results collected so far (used during calibration)
   def getResults(self):
      ret=self.results
      self.results=[]
      return ret

   # method to clear the results list
   def clearResults(self):
      self.results=[]
      return

   # the method called by the thread superclass to run the main loop of the thread.
   def run(self):
      # main loop of thread.
      print "Starting port reader."
      fhan=serial.Serial(port=self.usbPort, baudrate=usbBaud, timeout=0.1)
      fhan.open()
      seq=PySeq() # the sequencer library object
      port=seq.createOutPort('DrumMapper') # make a midi out port.
      while True:
         inbuf=[]
         while True:
            char=fhan.read(1)
            if self.quit: 
               fhan.close()
               return
            if len(char)==0: continue
            byte=ord(char) # convert to a byte. 
            #print "%02X " % byte,
            if byte>0xff:
              print "warning - bad character" # just in case.
            inbuf.append(byte & 0x7f) # put the last 7 bits in a buffer.
            if (byte & 0x80) != 0:
               #print " mark "
               if len(inbuf) >= 8: break # have found a complete set of values
               else: # there is a marker but not enough values to make a full set.
                  print "short buffer on usb read"
                  inbuf=[]
                  continue
         # should now have 8 * 7 bit values to make 4 * 14 bit sensor readings from.
         sens=[]
         for i in xrange(4):
            sens.append((inbuf.pop(0)<<7)+inbuf.pop(0))
         # have got the sensor values, so do something with them.
         vel=sum(sens); # velocity is the sum of all the sensor values.
         if self.mode == MODE_STOP: 
            print "Raw velocity:",vel,"readings:",sens
            continue
         if self.mode == MODE_MIDI:
            midiCoords=[]
            xyCoords=[]
            for coord in xrange(2): # i.e. do X coord then Y
               xyVal=mapCurve(calibrator.coeffs[coord],sens[0],sens[1],sens[2],sens[3])
               midiVal=int((xyVal+1)*64) # convert coord into 0..127 range for midi
               # do some range checking.
               if midiVal > 127: midiVal=127
               if midiVal < 0 : midiVal=0
               midiCoords.append(midiVal) # store the result.
               xyCoords.append(xyVal)
            # work out corrected velocity from velCurve
            cVel=velCurve(calibrator.velCoeffs,sum(sens),xyCoords[0],xyCoords[1])
            #print "corrected velocity:",cVel
            velOut=int(velOutRange[0]+(cVel*(velOutRange[1]-velOutRange[0])) )
            if velOut < 0 : velOut=0
            if velOut > 127 : velOut=127
            print "midi vel:",velOut,"; coords:",midiCoords, "; raw vel:",vel,"; readings:",sens
            # send results over midi port
            evt=snd_seq_event()
            evt.setController(midich,70,midiCoords[0])
            evt.sendNow(seq,port)
            evt=snd_seq_event()
            evt.setController(midich,71,midiCoords[1])
            evt.sendNow(seq,port)
            evt=snd_seq_event()
            evt.setNoteOn(midich,midiCoords[0],velOut)
            evt.sendNow(seq,port)
            evt=snd_seq_event()
            evt.setNoteOff(midich,midiCoords[0],velOut)
            evt.sendNow(seq,port)
            continue
         if self.mode == MODE_CAL: # in calibration mode.
            print "raw velocity:",vel, "raw sensor readings:",sens
            self.results.append(sens) # store results to be collected later.
            continue
         if self.mode == MODE_QUIT: # quit sent so end the thread.
            break

# The class which deals with calibration.
class Calibrator():
   def __init__(self):
      self.coeffs=[[],[]] # currently stored calibration coefficients
      self.velCoeffs=[] # range of raw velocities, min,max
      self.calLocs=[ ["top left",[1,1]], ["top centre",[1,0]], ["top right",[1,-1]], \
         ["top left mid",[0.5,0.5]], ["top right mid",[0.5,-0.5]], \
         ["left centre",[0,1]],["centre",[0,0]],["right centre",[0,-1]], \
         ["bottom left mid",[-0.5,0.5]],["bottom right mid",[-0.5,-0.5]], \
         ["bottom left",[-1,1]],["bottom centre",[-1,0]],["bottom right",[-1,-1]], \
         ] # series of locations where the user is asked to hit the pad during calibration, 
           # with names and x-y coords (on scale -1..+1 )
      self.numCalLocs=len(self.calLocs) # number of calibration locations
      self.calOK=False # is there some calibration data to work with?
      self.coeffsOK=False # are the calibration coefficients valid?
      self.inputHandle="" # file handle to read text input from.
      self.numCoeffs=7 # number of coefficients for first curve fit.
      self.numVelCoeffs=12 # number of coefficients for velocity mapping.
      self.velRange=[] # range of velocities TODO - remove this.
      self.calResults=[] # stores the raw results from a calibration.
   
   # load some previously saved calibration coefficients from a file.
   def load(self):
      print "Loading calibration data... ",
      calfHan=open(calFile,'r')
      self.coeffs=[0,0] # array to store x,y coefficients in.
      for coord in xrange(2): # do x coord first then y.
         line=calfHan.readline().strip()
         coordCoeffs=map(lambda x: float(x), line.split(", ")) # make the values numerical.
         self.coeffs[coord]=coordCoeffs # and store them in the x,y array.
      line=calfHan.readline().strip()
      velCoeffs=map(lambda x: float(x), line.split(", "))
      self.velCoeffs=velCoeffs
      calfHan.close()
      self.coeffsOK=True
      print "done"
      return

   # save calculated calibration coefficients to a file.
   def save(self):
      print "Saving calibration data... ",
      calfHan=open(calFile,'w')
      if self.coeffsOK:
         for c in self.coeffs:
            cStr=map(lambda x: str(x),c)
            cOut=", ".join(cStr)
            calfHan.write(cOut+"\n")
         vStr=map(lambda x: str(x), self.velCoeffs)
         vOut=", ".join(vStr)
         calfHan.write(vOut+"\n")
         calfHan.close()
         print "done"
      else:
         print "No calibration data to save"
      return

   # save actual results from sensors during calibration
   def saveRaw(self):
      print "Saving raw calibration data... ",
      calfHan=open(calDumpFile,'w')
      if self.calOK:
         for loc in xrange(len(self.calLocs)):
            locCoords=self.calLocs[loc][1]
            for res in self.calResults[loc]:
               resOut=[str(loc)]
               for sens in xrange(4):
                  resOut.append(str(res[sens]))
               strOut=", ".join(resOut)+"\n"
               calfHan.write(strOut)
         calfHan.close()
         print "Saved raw calibration data"
      else:
         print "No calibration data to save"
      return

   # load some previously saved raw calibration data.
   def loadRaw(self):
      print "Loading raw calibration data...",
      calfHan=open(calDumpFile,"r")
      self.calResults=[]
      for loc in xrange(len(self.calLocs)):
         self.calResults.append([])
      for line in calfHan:
         print line
         line=line.strip()
         vals=line.split(", ")
         vals=map(lambda x: int(x),vals)
         loc=vals[0] # first value is the location number.
         sens=vals[1:5] # last four are the sensor values.
         self.calResults[loc].append(sens) # put the data into calResults as if it had just come from a calibration.
      self.calOK=True
      print self.calResults
      print "done."

   # take calibration readings from sensors in each location.
   def measure(self,inHan):
      print "Measuring calibration data"
      self.calResults=map(lambda x: 0.0, self.calLocs) # same number of entries as calLocs but filled with zeroes
      portReader.clearResults() # clear any results from previous calibration.
      self.calOK=False # invalidate the previous calibration data.
      calLoc=0 # current calibration location (index into self.calLocs[])
      coeffsTmp=[0,0] # temp var for coefficients
      portReader.mode=MODE_CAL # switch the reader into calibration mode.
      while True:
         # prompt the user
         print "Tap the pad several times at the",self.calLocs[calLoc][0]," then type 'n'"
         print "(or 'r' to repeat a reading, or 'stop' to cancel calibration)."
         print "You should hit the pad at least 3 times each at the hardest and softest"
         print "velocity you want the pad to respond to, and a few times in between."
         line=inHan.readline().strip()
         # interpret the command typed
         if line == "stop":
            print "Calibration stopped."
            portReader.mode=MODE_STOP
            break
         if line == "n": # go to next location.
            self.calResults[calLoc]=portReader.getResults() # save results from last locn.
            portReader.clearResults()
            calLoc+=1 # move to next one.
            if calLoc==self.numCalLocs: # if it's the last one...
               self.calOK=True # now we have some valid data.
               break
         if line == "r": # repeat last set of readings.
            print "Please try again."
            portReader.clearResults()
      print "Finished measuring calibration data."
      return

   # calculate a set of coeeficients from some recent or saved calibration data.
   def calculate(self):
      global iters
      coeffsTmp=[[],[]]
      if not self.calOK:
         print "No Raw calibration data"
         return
      portReader.mode=MODE_STOP
      self.coeffsOK=False # invalidate the previous coefficients.
      self.velRange=[300,1000]
      print "Calculating coefficients for position mapping..."
      # calculate the coefficients for the x coord then the y coord.
      mvals=[[],[]] # an array of arrays of actual x and y coords.
      # DONE - rewritten this next bit so it merges the results from all the locations here
      # rather than when storing the data.
      for coord in xrange(2):
         print "   coordinate ",coord,":"
         sensVals=[[],[],[],[]] # an array of 1D arrays of sensor readings for each strike.
         for loc in xrange(len(self.calLocs)):
            locCoords=self.calLocs[loc][1]
            for res in self.calResults[loc]:
               mvals[coord].append(locCoords[coord])
               for sens in xrange(4):
                  sensVals[sens].append(res[sens]) 
         mvals[coord]=array(mvals[coord]) # convert to numpy style numerical array.
         for sens in xrange(4):
            sensVals[sens]=array(sensVals[sens]) # ditto.
         iters=0
         p0=zeros(self.numCoeffs)+1.0 # p0 is a first try at a set of coefficients for the equation.
         # now feed the rearranged results into the leastsq function.
         coeffsTmp[coord],cov_x,infodict,mesg,ier=leastsq(mapCurveErr,p0,args=(mvals[coord],sensVals[0],sensVals[1],sensVals[2],sensVals[3]),full_output=True,maxfev=10000)
         if ier<=4:
            print "      found least squares fit in",infodict['nfev'],"iterations"
         else:
            print "** Unable to fit curve:",mesg,"**"
            return
      print "Calculating coefficients for velocity mapping..."
      # fit the velocity mapping equation.
      cvVals=[] # an array of corrected velocities
      rvVals=[] # an array of raw velocities
      coordVals=[[],[]] # an array of two arrays of coordinate values.
      for loc in xrange(len(self.calLocs)):
         # first work out the velocity scaling for this location.
         locResults=sorted(self.calResults[loc],lambda r1,r2: cmp(sum(r1),sum(r2)))
         #print "locResults:",locResults 
         for res in xrange(len(locResults)):
            for sens in xrange(4):
               locResults[res][sens]=float(locResults[res][sens])
         velRange=[sum(locResults[0:3])/3, # ave of lowest 3 values.
                   sum(locResults[-3:])/3] # ave of highest 3 values.
         #print "velrange: ",velRange
         for res in self.calResults[loc]:
            rv=sum(res) # raw velocity is sum of sensor values.
            rvVals.append(rv) 
            cv=(float(rv)-velRange[0])/(velRange[1]-velRange[0]) # scale the corrected vel to [0,1]
            cvVals.append(cv)
            xy=[0,0]
            for coord in xrange(2):
               xy[coord]=mapCurve(coeffsTmp[coord],res[0],res[1],res[2],res[3])
               coordVals[coord].append(xy[coord])
            #print "rv:",rv,"x:",xy[0],"y:",xy[1],"cv:",cv           
      cvVals=array(cvVals)
      rvVals=array(rvVals)
      for coord in xrange(2):
         coordVals[coord]=array(coordVals[coord])
      iters=0
      p0=zeros(self.numVelCoeffs)+1.0 # p0 is a first try at a set of coefficients for the equation.
      # now feed the rearranged results into the leastsq function.
      velCoeffsTmp,cov_x,infodict,mesg,ier=leastsq(velCurveErr,p0,args=(cvVals,rvVals,coordVals[0],coordVals[1]),full_output=True,maxfev=10000)
      #print velCoeffsTmp
      if ier<=4:
         print "      found least squares fit in",infodict['nfev'],"iterations"
      else:
         print "** Unable to fit curve:",mesg,"**"
         return
      print "Finished calibration."
      self.velCoeffs=velCoeffsTmp # store the results
      self.coeffs=coeffsTmp
      self.coeffsOK=True # validate the calibration.
      return

# main command loop reads from stdin, accepts commands and acts on them.
def commandLoop(): 
   inhan=open("/dev/stdin",'r')
   while True:
      line=inhan.readline().strip()
      if line=="midi":
         if calibrator.coeffsOK:
            print "Now sending midi events"
            portReader.mode=MODE_MIDI
         else:
            print "Pad is not calibrated. Can't enter midi mode."
      elif line=="stop":
         print "Midi events stopped"
         portReader.mode=MODE_STOP
      elif line=="save":
         calibrator.save()
      elif line=="saveraw":
         calibrator.saveRaw()
      elif line=="calib":
         calibrator.measure(inhan)
         calibrator.calculate()
      elif line=="calfile":
         calibrator.loadRaw()
         calibrator.calculate()
      elif line=="load":
         calibrator.load()
      elif line=="quit":
         print "Quitting midimapper."
         portReader.mode=MODE_QUIT
         break
      else:
         print "No such command. Commands are midi, stop, save, saveraw, calib, load, calfile, quit"

# the main program
portReader=PortReader(usbPort) # create a thread object to read from the usb port
calibrator=Calibrator() # create a calibration object.
portReader.start() # start the read thread.
commandLoop() # start the main command loop.
portReader.quit=True # tell portReader thread to quit.
portReader.join() # wait for the reader thread.
print "bye."
